//
// Created by q1w2e3r4 on 23-7-25.
//

#include <queue>
#include "HIR-opt/HOptimizer.h"
#include "HIR-opt/DomTreePass.h"
#include "HIR-opt/LiveVariableAnalysis.h"
#include "HIR-opt/Mem2RegPass.h"
#include "HIR-opt/SccpPass.h"
#include "HIR-opt/RenamePass.h"
#include "HIR-opt/DePhiPass.h"
#include "HIR-opt/LoopPass.h"

HOptimizer::HOptimizer(GlobalUnit *gu) {
    this->globalUnit = gu;
}

void HOptimizer::Optimize() {
    // Dom Tree & DF
    BeforeOpt();

    //globalUnit->Emit(std::cerr);
    DomTreePass* pass1 = new DomTreePass(this->globalUnit);
    pass1->run();

    //LVA
    LiveVariableAnalysis* lva = new LiveVariableAnalysis(this->globalUnit);

    lva->analysis();
//
//     Mem2reg
    Mem2RegPass* mem2reg = new Mem2RegPass(this->globalUnit);
    mem2reg->run();

//  Loop opt
//    LoopPass * loopPass = new LoopPass(this->globalUnit);
//    loopPass->run();
//  SccpPass
    SccpPass* sccp = new SccpPass(this->globalUnit);
    sccp->run();
//
    // Rename
    RenamePass* renamePass = new RenamePass(this->globalUnit);
    renamePass->run();
////
//    globalUnit->Emit(std::cerr);
    DePhiPass* dePhiPass = new DePhiPass(this->globalUnit);
    dePhiPass->run();

    //cerr << "after opt:" << endl;
}

void HOptimizer::BeforeOpt() {
    // 1.bfs && remove unvisited blocks
    for(auto&[name,func]: globalUnit->func_table){
        if(func->entry == nullptr) continue;

        set<BasicBlock*> vis;
        queue<BasicBlock*> q;
        q.push(func->entry);
        while(!q.empty()){
            BasicBlock * block = q.front(); q.pop();
            if(vis.count(block)) continue;
            vis.insert(block);
            for(auto bb:block->succ){
                q.push(bb);
            }
        }

        auto& v = func->block_list;
        vector<BasicBlock *> not_visited;
        for(auto & it : v){
            if(!vis.count(it)){ // not visited
                not_visited.push_back(it);
            }
        }

        for(auto bb:not_visited){
            DelBlock(bb);
        }
    }
}
